/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is WebDavBrowser.
 *
 * The Initial Developer of the Original Code is
 *   Rafa Malinowski, Pozna Supercomputing And Networing Center
 * Portions created by the Initial Developer are Copyright (C) 2007
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

const nsITreeView = Components.interfaces.nsITreeView;

/**
 * nsITreeView implementation for WebDavBrowsers.
 * Uses webDavTreeItem for displaing nsIWebDavResources.
 */
WebDavTreeView.prototype = {
	_xul: null, // xul element of treeview
	_items: null,
	_selection: null,
	_tree: null,
	_sortColumn: "name", // current sort column
	_sortIncreasing: true,

	get rowCount() { return this._items.length; },
	get selection() { return this._selection; },
	set selection(selection) { this._selection = selection; },
	get tree() { return this._tree; },
	
	/**
	 * Add one item at level and index
	 */
	addItem: function (level, index, item) {
		item.setSorting(this._sortColumn, this._sortIncreasing);
		this._items.splice(index, 0, item);
		this._updateIndexes(index);
		this.tree.rowCountChanged(index, 1);
		item.addedToTreeView(this, level, index);
	},
	
	/**
	 * Add many items at level and first at index
	 */
	addItems: function(level, index, items) {
		var insertionIndex = index;
	
		for (var i = 0; i < items.length; i++) {
			items[i].setSorting(this._sortColumn, this._sortIncreasing);
			this._items.splice(insertionIndex, 0, items[i]);
			items[i].addedToTreeView(this, level, insertionIndex);
			insertionIndex++;
		}
		
		this._updateIndexes(index);
		this._tree.rowCountChanged(index, items.length);
	},
	
	/**
	 * Remove item and all of its subitems.
	 */
	removeItem: function (item) {
		var count = item.subItemsCount + 1;

		this._items.splice(item.index, count);
		this._updateIndexes(item.index - 1);
		this._tree.rowCountChanged(item.index, -count);
	},

	/**
	 * Remove count items from index. Dont care about subitems.
	 * Them shoul be included in 'count' variable. If not - it is a bug.
	 */
	removeItems: function (index, count) {
		this._items.splice(index, count);
		this._updateIndexes(index);
		this._tree.rowCountChanged(index, -count);
	},

	/**
	 * Adds item at top of the tree.
	 */
	addRootItem: function (item) {
		this.addItem(0, 0, item);
	},
	
	itemUpdated: function(row) {
		this._tree.invalidateRow(row);
	},

	/**
	 * Allows only drops on collection and before/after non-collections.
	 */
	canDrop: function (row, orientation) {
		if (row == -1)
			return false;
		if ((row == 0) && (orientation == nsITreeView.DROP_BEFORE))
			return false;
		return (this._items[row].isCollection || (orientation != nsITreeView.DROP_ON));
	},

	cycleCell: function (row, col) {
	},

	/**
	 * Sort items by given columnt.
	 */
	cycleHeader: function (col) {
		if (col.id == "tree-view-column-status")
			return;

		if ("tree-view-column-" + this._sortColumn == col.id) {
			this._sortIncreasing = !this._sortIncreasing;
			if (this._sortIncreasing)
				this._getXulColumn(this._sortColumn).setAttribute("sortDirection", "ascending");
			else
				this._getXulColumn(this._sortColumn).setAttribute("sortDirection", "descending");
		} else {
			this._sortIncreasing = true;
			this._getXulColumn(this._sortColumn).removeAttribute("sortActive");
			this._getXulColumn(this._sortColumn).removeAttribute("sortDirection");
			this._sortColumn = col.id.substr(17);
			this._getXulColumn(this._sortColumn).setAttribute("sortActive", true);
			this._getXulColumn(this._sortColumn).setAttribute("sortDirection", "ascending");
		}
		
		for (var i in this._items)
			this._items[i].setSorting(this._sortColumn, this._sortIncreasing);
		
		var rootItem = this._items[0];
		this._tree.rowCountChanged(0, -this._items.length);
		this._items = new Array();
		
		this._fillTree(rootItem);
		
		this._updateIndexes();
		this._tree.rowCountChanged(0, this._items.length);
		this._tree.invalidate();
	},
	
	/**
	 * Adds an item and its subitems (and its subiems and so on) to the end of tree.
	 */
	_fillTree: function(item) {
		this._items[this._items.length] = item;
		var items = item.items;
		for (var i in items)
			this._fillTree(items[i]);
	},
	
	_handleSelfDrop: function(destination, move) {
		var selection = webDavBrowser._webDavTreeView.selection;
		var rangeCount = selection.getRangeCount();
		var min = {};
		var max = {};
		var urls = new Array();
		var index = 0;

		for (var i = 0; i < rangeCount; i++) {
			selection.getRangeAt(i, min, max);
			for (var j = min.value; j <= max.value; j++) {
				switch (move) {
					case Components.interfaces.nsIDragService.DRAGDROP_ACTION_MOVE:
						if (this._items[j]._webDavResource.moveTo(destination, true, false) == null)
							alert("Unable to move resource: " + this._items[j]._webDavResource.path);
						break;
					case Components.interfaces.nsIDragService.DRAGDROP_ACTION_COPY:
						if (this._items[j]._webDavResource.copyTo(destination, true, false) == null)
							alert("Unable to copy resource: " + this._items[j]._webDavResource.path);
						break;
				}
			}
		}

	},
	
	_copyFileToWebDav: function(destination, source) {
		if (destination.loaded != Components.interfaces.nsIWebDavResource.WEBDAV_LOADED_LOADED) {
			// error message
			return;
		}
		
		var resource = destination.createSubResource(source.leafName, false);
		if (resource != null) // async create file
			resource.createResourceFromFile(source, false);
	},
	
	_copyDirectoryToWebDav: function(destination, source) {
		if (destination.loaded != Components.interfaces.nsIWebDavResource.WEBDAV_LOADED_LOADED) {
			// error message
			return;
		}

		var destinationCollection = destination.createSubResource(source.leafName, true);
		if (destinationCollection == null) {
			// error message
			return;
		}

		destinationCollection.createCollection(true);
		
		if (destinationCollection == null) {
			// error message
			return;
		}

		if (destinationCollection.loaded != Components.interfaces.nsIWebDavResource.WEBDAV_LOADED_LOADED) {
			// error message
			return;
		}
		
		var entries = source.directoryEntries;
		while (entries.hasMoreElements()) {
			var subElement = entries.getNext().QueryInterface(Components.interfaces.nsIFile);
			if (subElement == null)
				continue;
			if (subElement.isDirectory())
				this._copyDirectoryToWebDav(destinationCollection, subElement);
			else
				this._copyFileToWebDav(destinationCollection, subElement);
		}
	},

	/**
	  Get drops of a file on tree view.
	 */
	drop: function (row, orientation) {
		if (row == -1)
			return false;

		var dragService = Components.classes["@mozilla.org/widget/dragservice;1"]
			.getService(Components.interfaces.nsIDragService);
		var dragSession = dragService.getCurrentSession();
		
		// on windows it is a file
		var dataFlavor = "application/x-moz-file";
		if (!dragSession.isDataFlavorSupported(dataFlavor)) {
			// on linux it is a file-name
			dataFlavor = "text/unicode";
			if (!dragSession.isDataFlavorSupported(dataFlavor)) {
				dataFlavor = "moz/web-dav-drag-prompt";
				if (!dragSession.isDataFlavorSupported(dataFlavor))
					return;
			}
		}

		// select collection to add this item to
		// either current (dropped on) or parent collection
		var addToCollection = null;	
		if (orientation == nsITreeView.DROP_ON) // drop on collection = select current
			if (this._items[row].isCollection)
				addToCollection = this._items[row];
			else
				return;
		else
			if (row == 0) // drop on root = select root
				addToCollection = this._items[row];
			else // drop after collection = select current
				if ((orientation == nsITreeView.DROP_AFTER) && (this._items[row].isCollection))
					addToCollection = this._items[row];
				else // select parent
					addToCollection = this._items[row].parent;

		if (!addToCollection)
			return;

	    if (dataFlavor == "moz/web-dav-drag-prompt") {
	    	this._handleSelfDrop(addToCollection._webDavResource, dragSession.dragAction);
	    	return;
	    }

		var transferObject = Components.classes["@mozilla.org/widget/transferable;1"]
			.createInstance();
		transferObject = transferObject.QueryInterface(Components.interfaces.nsITransferable);
		transferObject.addDataFlavor(dataFlavor);

		var numItems = dragSession.numDropItems;
		for (var i = 0; i < numItems; i++) {
			dragSession.getData(transferObject, i);

    		var dataObj = new Object();
		    var dropSizeObj = new Object();

		    transferObject.getTransferData(dataFlavor, dataObj, dropSizeObj);
		    
		    var droppedFile = null;
		    if (dataFlavor == "text/unicode") { // on unix create file from file name
		    	var filePath = "";

				if ("nsISupportsWString" in Components.interfaces)
					filePath = new String(dataObj.value.QueryInterface(Components.interfaces.nsISupportsWString));
				else if ("nsISupportsString" in Components.interfaces)
					filePath = new String(dataObj.value.QueryInterface(Components.interfaces.nsISupportsString));
				else
					continue;

				if (filePath.substr(0, 7) == "file://")
					filePath = filePath.substr(7);
				else
					continue;

				droppedFile = Components.classes["@mozilla.org/file/local;1"]
					.createInstance();
				droppedFile = droppedFile.QueryInterface(Components.interfaces.nsILocalFile);
				droppedFile.initWithPath(filePath);
		    } else if (dataFlavor == "application/x-moz-file") {
			    droppedFile = dataObj.value.QueryInterface(Components.interfaces.nsILocalFile);
		    }

			if (droppedFile.isDirectory()) {
				this._copyDirectoryToWebDav(addToCollection.webDavResource, droppedFile);
				return;
			} else if (!droppedFile.isFile())
				return;

			if (!addToCollection.webDavResource.isNameValid(droppedFile.leafName)) {
				alert("Server does not support file name: " + droppedFile.leafName);
				continue;
			}

		    var webDavResource = addToCollection.webDavResource.createSubResource(droppedFile.leafName, false);
			if (webDavResource != null) // async create file
				webDavResource.createResourceFromFile(droppedFile, false);
			else // oops, file name invalid
				alert("Unable to add file: " + filePath);
		}

		// display the item by opening collection it was added to
		addToCollection.open();
	},

	getCellProperties: function (row, col, properties) {
	},

	getCellText: function (row, col) {
		if (col == this._getColumn("name")) {
			return this._items[row].displayName;
		} else if (col == this._getColumn("size")) {
			return this._items[row].size;
		} else if (col == this._getColumn("type")) {
			return this._items[row].mimeType;
		} else if (col == this._getColumn("creation-date")) {
			return this._items[row].creationDate;
		} else if (col == this._getColumn("modification-date")) {
			return this._items[row].modificationDate;
		} else if (col == this._getColumn("status")) {
			return this._items[row].status;
		} else
			return "";
	},

	getCellValue: function (row, col) {
		return "";
	},

	getColumnProperties: function (col, properties) {
	},

	getImageSrc: function (row, col) {
		if (col != this._getColumn("name"))
			return;

		// TODO: images stolen from gopher directory, needs new one
		if (this._items[row].isCollection)
			return "chrome://webdavbrowser/content/img/collection.png";
		else
			return "chrome://webdavbrowser/content/img/file.png";
	},

	getLevel: function (row) {
		return this._items[row].level;
	},

	getParentIndex: function (row) {
		var parent = this._items[row].parent;

		if (parent)
			return parent.index;
		else
			return -1;
	},

	getProgressMode: function (row, col) {
	},
	
	getRowProperties: function (index, properties) {
	},

	hasNextSibling: function (row, afterIndex) {
		var parent = this._items[row].parent;
			
		if (parent)
			return afterIndex < parent.lastItem.index;
		else
			return false;
	},

	isContainer: function (row) {
		return this._items[row].isCollection;
	},

	isContainerEmpty: function (row) {
		return this._items[row].isCollectionEmpty;
	},

	isContainerOpen: function (row) {
		return this._items[row].opened;
	},

	isEditable: function (row, col) {
		return false;
	},

	isSeparator: function (index) {
		return false;
	},

	isSorted: function () {
		return true;
	},

	performAction: function () {
	},

	performActionOnCell: function () {
	},

	performActionOnRow: function (action, row) {
	},

	selectonChanged: function () {
	},

	setCellText: function () {
	},

	setCellValue: function () {
	},

	setTree: function (tree) {
		this._tree = tree;
	},

	toggleOpenState: function (row) {
		this._items[row].toggleOpenState();
	},

	/**
	 * Gives every TreeItem and index so it can be easily identified.
	 */
	_updateIndexes: function (start) {			
		for (var i = start; i < this._items.length; i++)
			this._items[i].index = i;
		this._tree.invalidate();
	},
	
	onTreeDblClick: function(x, y) {
		var row = this._tree.getRowAt(x, y);
		
		if (row != -1)
			if (!this._items[row].isCollection)
				this._items[row].open();
	},

	_getColumn: function(name) {
		return this._xul.columns.getNamedColumn("tree-view-column-" + name);
	},

	_getXulColumn: function(name) {
		return document.getElementById("tree-view-column-" + name);
	},
	
	onResourceMoveFinished: function(host, path) {
		// TODO: write content
		alert("onResourceMoveFinished(" + host + ", " + path +")");
	}
};

function WebDavTreeView(document) {
	this._xul = document.getElementById("webdav-browser-tree");
	this._items = new Array();
	return this;
}
